////////////////////////////////////////////////////////////////////////
// Class:       DataDumpHDFT962
// Plugin Type: analyzer (art v3_03_01)
// File:        DataDumpHDFT962_module.cc
//
// Generated at Tue Nov 19 13:37:52 2019 by Tingjun Yang using cetskelgen
// from cetlib version v3_08_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art_root_io/TFileService.h"

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardataobj/RawData/raw.h" //raw::Uncompress()
#include "lardataobj/RawData/RawDigit.h"
#include "larcore/Geometry/Geometry.h"

#include "TTree.h"
#include "TTimeStamp.h"

#include <fstream>
#include <vector>
#include <algorithm>

#include "hep_hpc/hdf5/File.hpp"
#include "hep_hpc/hdf5/Ntuple.hpp"

using wire_nt_t = hep_hpc::hdf5::Ntuple<float>;
using evt_nt_t = hep_hpc::hdf5::Ntuple<unsigned int, double>;

inline std::array<unsigned int, 3>
get_eid(art::Event const& e)
{
  return { e.run(), e.subRun(), e.id().event() };
}

bool chIncrease(art::Ptr<recob::Wire> const & w1,
                art::Ptr<recob::Wire> const & w2){
  return (w1->Channel()<w2->Channel());
}

bool rawChIncrease(art::Ptr<raw::RawDigit> const & rd1,
                   art::Ptr<raw::RawDigit> const & rd2) {
  return (rd1->Channel() < rd2->Channel());
}


namespace t962 {
  class DataDumpHDFT962;
}

class t962::DataDumpHDFT962 : public art::EDAnalyzer {
public:
  explicit DataDumpHDFT962(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  DataDumpHDFT962(DataDumpHDFT962 const&) = delete;
  DataDumpHDFT962(DataDumpHDFT962&&) = delete;
  DataDumpHDFT962& operator=(DataDumpHDFT962 const&) = delete;
  DataDumpHDFT962& operator=(DataDumpHDFT962&&) = delete;

  // Required functions.
  void analyze(art::Event const& e) noexcept;
  void beginJob();
  virtual ~DataDumpHDFT962() noexcept;


protected:

//geo::GeometryCore const* fGeom;
  //TTree *fTree;
  double evttime;
//  std::vector<unsigned short> channel;
//  std::vector<unsigned short> tick;
//  std::vector<float> adc;
  hep_hpc::hdf5::File hdffile;

private:
  std::string fRawDigitLabel;
  art::InputTag fWireProducerLabel;

};


t962::DataDumpHDFT962::DataDumpHDFT962(fhicl::ParameterSet const& p)
  : EDAnalyzer{p}  // ,
  // More initializers here.
{
  // Call appropriate consumes<>() for any products to be retrieved by this module.
  fRawDigitLabel             = p.get<std::string>("RawDigitLabel", "");
  fWireProducerLabel         = p.get<art::InputTag>("WireProducerLabel", "caldata");

  if (fRawDigitLabel.empty() && fWireProducerLabel.empty()){
    throw cet::exception("DataDumpHDFT962") << "Both RawDigitLabel and WireProducerLabel are empty";
  }

  if ((!fRawDigitLabel.empty()) && (!fWireProducerLabel.empty())){
    throw cet::exception("DataDumpHDFT962") << "Only one of RawDigitLabel and WireProducerLabel should be set";
  }
}
t962::DataDumpHDFT962::~DataDumpHDFT962() noexcept
{
}


void t962::DataDumpHDFT962::analyze(art::Event const& e) noexcept
{

//fGeom = &*(art::ServiceHandle<geo::Geometry>());

  hdffile = hep_hpc::hdf5::File(Form("r%d_e%d.h5",e.run(),e.event()), H5F_ACC_TRUNC);
  auto event_id = get_eid(e);

  art::Timestamp ts = e.time();
  if (ts.timeHigh() == 0){
    TTimeStamp tts(ts.timeLow());
    evttime = tts.AsDouble();
  }
  else{
    TTimeStamp tts(ts.timeHigh(), ts.timeLow());
    evttime = tts.AsDouble();
  }
//  channel.clear();
//  tick.clear();
//  adc.clear();

  //std::ofstream outfile (Form("r%de%d.txt",run,event));

  //wire_nt_t wiresigs(hdffile, "wiresigs", {{"eid",4}});
  wire_nt_t wiresigs(hdffile, "wiresigs", {"adc"});
  evt_nt_t evtids(hdffile, "evtids", {{"eid",3}, "evttime"});
  evtids.insert(event_id.data(), evttime);

  lariov::ChannelStatusProvider const& channelStatus = art::ServiceHandle<lariov::ChannelStatusService>()->GetProvider();

  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, "", rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }

  // git wire
  art::Handle < std::vector < recob::Wire > > wireListHandle;
  std::vector < art::Ptr < recob::Wire > > wires;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wires, wireListHandle);
  }

  int oldchannel = -1;
  // using wires
  if (wires.size()) {
    std::cout << "wires.size(): " << wires.size() << std::endl;
    std::sort(wires.begin(), wires.end(), chIncrease);
    for (auto & wire : wires){
      int channel = wire->Channel();
      if(channelStatus.IsBad(channel)) continue;
      std::cout<<"Channel = "<<channel<<std::endl;

      for (int i = oldchannel + 1; i<channel; ++i){
        for (int j = 0; j<2048; ++j){
          wiresigs.insert(0.);
        }
      }
      oldchannel = channel;

      const recob::Wire::RegionsOfInterest_t& signalROI = wire->SignalROI();
      int lasttick = 0;
      for(const auto& range : signalROI.get_ranges()){
        const auto& waveform = range.data();
        // ROI start time
        raw::TDCtick_t roiFirstBinTick = range.begin_index();

        for (int i = lasttick; i<roiFirstBinTick; ++i){
          wiresigs.insert(0.);
          ++lasttick;
        }
        for(size_t idx = 0; idx < waveform.size(); idx++){
          wiresigs.insert(waveform[idx]);
          ++lasttick;
        }
      }
      for (int i = lasttick; i<2048; ++i){
        wiresigs.insert(0.);
      }
    } // end of loop wires
  } // recob::Wire
  else if (rawdigitlist.size()) {
    std::cout << "rawdigitlist.size(): " << rawdigitlist.size() << std::endl;
    std::sort(rawdigitlist.begin(), rawdigitlist.end(), rawChIncrease);
    for (auto & rawdigit : rawdigitlist){
      int channel = rawdigit->Channel();
      if(channelStatus.IsBad(channel)) continue;
      std::cout<<"Channel = "<<channel<<std::endl;

      for (int i = oldchannel + 1; i<channel; ++i){
        for (int j = 0; j<2048; ++j){
          wiresigs.insert(0.);
        }
      }
      oldchannel = channel;

      std::vector<float> adcvec(2048);
      raw::RawDigit::ADCvector_t rawadc(2048);
      raw::Uncompress(rawdigit->ADCs(), rawadc, rawdigit->Compression());

      float pedestal = rawdigit->GetPedestal();
      for (size_t j=0; j < rawadc.size(); ++j) {
        adcvec[j] = rawadc[j] - pedestal;
      }
      
      for (int j=0; j<2048; j++) {
        wiresigs.insert(adcvec[j]);
      }
    } // end of loop rawdigitlist
  } // RawDigit

  for (int i = oldchannel + 1; i<480; ++i){
    for (int j = 0; j<2048; ++j){
      wiresigs.insert(0.);
    }
  }

//    int nticks = wire->Signal().size();
//    for (int j = 0; j < nticks; j++){
//      float adc = wire->Signal()[j];
//      wiresigs.insert(adc);
//    }
//    if (nticks <6000){
//      for (int j = nticks; j < 6000; j++)
//        wiresigs.insert(0.);
//    }

  std::cout<<"event_time: "<<evttime<<std::endl;
  
}

void t962::DataDumpHDFT962::beginJob()
{
}

DEFINE_ART_MODULE(t962::DataDumpHDFT962)
