////////////////////////////////////////////////////////////////////////
// Class:       CNNRoiAna
// Plugin Type: analyzer (art v3_05_01)
// File:        CNNRoiAna_module.cc
//
// Generated at Mon Oct  5 20:07:45 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
// 
// About: analysis the prediction score of cnn roi
//   (a) default cnn roi window size 200 ticks with stride size 150 ticks
//   (b) input contains the cnn roi prediction scores, not rois
//   (c) because stide size is different from window size, the prediction
//       scores are grouped every 150 ticks, except the last one that has
//       98 ticks
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art_root_io/TFileService.h"
#include "art/Utilities/make_tool.h"

#include "larcore/Geometry/Geometry.h"
#include "lardata/DetectorInfoServices/DetectorClocksService.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"

#include "TH1D.h"

using namespace std;

class CNNRoiAna;


class CNNRoiAna : public art::EDAnalyzer {
public:
  explicit CNNRoiAna(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  CNNRoiAna(CNNRoiAna const&) = delete;
  CNNRoiAna(CNNRoiAna&&) = delete;
  CNNRoiAna& operator=(CNNRoiAna const&) = delete;
  CNNRoiAna& operator=(CNNRoiAna&&) = delete;

  // Required functions.
  void analyze(art::Event const& e) override;

private:

  // User's function
  void beginJob() override;
  void endJob() override;
  bool isSignalInROI(int maxtick, int roistart, int roiend);
  void getSignal(const sim::SimChannel& simchannel, std::vector<int> & signal_starttick, std::vector<int> & signal_endtick, std::vector<double> & signal_max, std::vector<int> & signal_max_tdctick);

  // fhicl parameters
  art::InputTag fWireProducerLabel;
  art::InputTag fSimulationProducerLabel;
  bool fUseNumElectrons;
  double fSignalNumElectronsLowCut;
  double fSignalNumElectronsHighCut;
  bool fSaveSignal;
  bool fSaveNoise;
  int fWindowSize;
  int fStrideSize;
  bool fDropLastWindow;
  
  // user's parameters
  int fNPlanes;
  int fNTicks;

  TH1D* h1_cnnroi_signal[3];
  TH1D* h1_cnnroi_noise_pure[3];

};


CNNRoiAna::CNNRoiAna(fhicl::ParameterSet const& p)
  : EDAnalyzer{p}  // ,
  // More initializers here.
{
  fWireProducerLabel         = p.get<art::InputTag>("WireProducerLabel");
  fSimulationProducerLabel   = p.get<art::InputTag>("SimulationProducerLabel","largeant");
  fUseNumElectrons           = p.get<bool>("UseNumElectrons", true);
  fSignalNumElectronsLowCut  = p.get<double>("SignalNumElectronsLowCut");
  fSignalNumElectronsHighCut = p.get<double>("SignalNumElectronsHighCut");
  fSaveSignal                = p.get<bool>("SaveSignal", true);
  fSaveNoise                 = p.get<bool>("SaveNoise", true);
  fWindowSize                = p.get<int>("WindowSize", 200);
  fStrideSize                = p.get<int>("StrideSize", 150);
  fDropLastWindow            = p.get<bool>("DropLastWindow", true);

  // user's parameters
  fNPlanes = lar::providerFrom<geo::Geometry>()->Nplanes();

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);
  fNTicks = detProp.NumberTimeSamples(); // number of clock ticks per event
}

void CNNRoiAna::analyze(art::Event const& e)
{
  auto const* geo = lar::providerFrom<geo::Geometry>();
  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataFor(e);
  auto const detProp =
    art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataFor(e, clockData);
  auto const& chStatus = art::ServiceHandle<lariov::ChannelStatusService>()->GetProvider();

  art::Handle<std::vector<recob::Wire>> wireListHandle;
  std::vector<art::Ptr<recob::Wire>> wirelist;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wirelist, wireListHandle);
  }

  auto simChannelHandle = e.getValidHandle<std::vector<sim::SimChannel>>(fSimulationProducerLabel);

  if (fSaveSignal) {
    // loop over simChannels
    for (auto const& simchannel : (*simChannelHandle)) {

      // .. get simChannel channel number
      const raw::ChannelID_t channel = simchannel.Channel();
      if (chStatus.IsBad(channel)) 
        continue;

      int view = geo->View(channel);

      std::vector<int> signal_starttick;
      std::vector<int> signal_endtick;
      std::vector<double> signal_max; // according to number of electrons or energy deposits
      std::vector<int> signal_max_tdctick;
      
      getSignal(simchannel, signal_starttick, signal_endtick, signal_max, signal_max_tdctick);

      if (signal_starttick.size() == 0 ||
          (signal_endtick.size() == 1 && signal_endtick.back() == -999))
        continue;
      
      // find cnn roi score for "signal"
      // recob wire
      for (size_t w = 0; w < wirelist.size(); w++) {
        if (wirelist[w]->Channel() != channel) continue;

        // cnn roi prediction score
        std::vector<float> vecScore = wirelist[w]->Signal();

        // loop over signals
        for (size_t s = 0; s < signal_starttick.size(); ++s) {
          if (signal_max[s] < fSignalNumElectronsLowCut) continue;
          if (fSignalNumElectronsHighCut != -1 && signal_max[s] > fSignalNumElectronsHighCut) continue;

          double maxsignal = signal_max[s];
          
          int strideindex = int(signal_max_tdctick[s]/fStrideSize);

          if (fDropLastWindow && strideindex == int(fNTicks/fStrideSize)) {
            continue;
          }

          // check the following signals
          for (size_t s2 = s + 1; s2 < signal_starttick.size(); ++s2) {
            if (signal_max[s2] < fSignalNumElectronsLowCut) continue;
            if (fSignalNumElectronsHighCut != -1 && signal_max[s2] > fSignalNumElectronsHighCut) continue;

            if (signal_max_tdctick[s2] >= strideindex*fStrideSize + fWindowSize) {
            //if (signal_max_tdctick[s2] >= strideindex*fStrideSize + fStrideSize) {
              s = s2 - 1;
              break;
            }
            else {
              if (s2 == signal_starttick.size() - 1) s = s2; 
              if (signal_max[s2] > maxsignal) maxsignal = signal_max[s2];
            }
          } // loop over signal s2
          
          // fill cnn score
          double score = vecScore[strideindex*fStrideSize+1];
          if (score >= 1) score = 1. - 1e-7;
          h1_cnnroi_signal[view]->Fill(score);

        } // loop over signals s
        
      } // loop over wire w

    } // loop over simChannels
  } // fSaveSignal

  if (fSaveNoise) {
    std::map<raw::ChannelID_t, bool> signalMap;
    for (auto const& channel : (*simChannelHandle)) {
      signalMap[channel.Channel()] = true;
    }
    
    for (size_t i=0; i<wirelist.size(); i++) {
      int channel = wirelist[i]->Channel();
      if (signalMap[channel]) continue;
      if (chStatus.IsBad(channel)) continue;

      int view = geo->View(channel);

      // cnn roi prediction score
      std::vector<float> vecScore = wirelist[i]->Signal();
     
      int nstrides = int(fNTicks/fStrideSize+1);
      if (fDropLastWindow) nstrides = nstrides - 1;

      // fill cnn score
      for (int k=0; k<nstrides; k++) {
        double score = vecScore[k*fStrideSize+1];
        if (score >= 1) score = 1. - 1e-7;
        h1_cnnroi_noise_pure[view]->Fill(score); 
      }
    
    } // loop over wirelist i
  
  } // fSaveNoisel
}


void CNNRoiAna::beginJob()
{
  art::ServiceHandle<art::TFileService const> tfs;

  for (int i=0; i<fNPlanes; ++i) {
    if (fSaveSignal) {
      h1_cnnroi_signal[i] = tfs->make<TH1D>(Form("h1_cnnroi_signal_%d", i), Form("Plane %d; cnn score; Count [#]", i), 1000, 0, 1);
      h1_cnnroi_noise_pure[i] = tfs->make<TH1D>(Form("h1_cnnroi_noise_pure_%d", i), Form("Plane %d; cnn score; Count [#]", i), 1000, 0, 1);

    }
  }
}


void CNNRoiAna::endJob()
{

}


bool CNNRoiAna::isSignalInROI(int maxtick, int roistart, int roiend)
{
  // roiend is excluded within the ROI: [roistart, roiend)
  return roistart <= maxtick && maxtick < roiend;
}


void CNNRoiAna::getSignal(const sim::SimChannel& simchannel, std::vector<int> & signal_starttick, std::vector<int> & signal_endtick, std::vector<double> & signal_max, std::vector<int> & signal_max_tdctick) {

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);

  auto const& timeSlices = simchannel.TDCIDEMap();

  // time slice from simChannel is for individual tick
  // group neighboring time slices into a "signal"
  int tdctick_previous = -999;
  double maxE = -999.; // number of electrons or energy deposits
  int maxEtick = -999;
  
  for (auto const& timeSlice : timeSlices) {
    auto const tpctime = timeSlice.first;
    auto const& energyDeposits = timeSlice.second;
    int tdctick = static_cast<int>(clockData.TPCTDC2Tick(double(tpctime)));
    if (tdctick < 0 || tdctick > int(detProp.ReadOutWindowSize()) - 1) continue;

    // for a time slice, there may exist more than one energy deposit.
    double tdc_signal = 0;
    for (auto const& energyDeposit : energyDeposits) {
      if (fUseNumElectrons) {
        tdc_signal += energyDeposit.numElectrons;
      }
      else {
        tdc_signal += energyDeposit.energy;
      }
    }

    if (tdctick_previous == -999) {
      signal_starttick.push_back(tdctick);
      maxE = tdc_signal;
      maxEtick = tdctick;
    }
    else if (tdctick - tdctick_previous != 1) {
      signal_starttick.push_back(tdctick);
      signal_endtick.push_back(tdctick_previous);
      signal_max.push_back(maxE);
      signal_max_tdctick.push_back(maxEtick);
      maxE = tdc_signal;
      maxEtick = tdctick;
    }
    else if (tdctick - tdctick_previous == 1) {
      if (maxE < tdc_signal) {
        maxE = tdc_signal;
        maxEtick = tdctick;
      }
    }

    tdctick_previous = tdctick;

  } // loop over timeSlices timeSlice

  signal_endtick.push_back(tdctick_previous); // for last one
  signal_max.push_back(maxE);          // for last one
  signal_max_tdctick.push_back(maxEtick);     // for last one
}


DEFINE_ART_MODULE(CNNRoiAna)
